#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2021 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


import xrayutilities as xu

from .source import Source
from .scan import Scan


class Spec(Source):
    """Spec

    Source implementation for SPEC files.

    Args:
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.

    Keyword Args:
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        read_all_data (bool): read all data on parsing.
          If false, data will be read only on demand.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    Attributes:
        log (logging.logger): logger instance from logging.
        name (str): name of the source
        scan_dict (dict(scan)): dict of scan objects with
          key being the scan number.
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        h5_file_exists(bool): if h5 file exists.
        read_all_data (bool): read all data on parsing.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    """
    def __init__(self, file_name, file_path, **kwargs):
        super().__init__(file_name, file_path, **kwargs)

    def parse_raw(self):
        """parse_raw

        Parse the raw source file/folder and populate the `scan_dict`.

        """
        self.log.info('parse_raw')

        if ('spec_file' not in dir(self)) or self.force_overwrite:
            self.log.info('Create spec_file from xrayutilities')
            self.spec_file = xu.io.SPECFile(self.file_name,
                                            path=self.file_path)

        # update the xu.spec_file
        self.spec_file.Update()
        # iterate through scan list in xu.spec_file
        for spec_scan in self.spec_file.scan_list:
            # check for scan number in given range
            if (spec_scan.nr >= self.start_scan_number) and \
                    ((spec_scan.nr <= self.stop_scan_number) or
                        (self.stop_scan_number == -1)):
                last_scan_number = self.get_last_scan_number()
                # check if Scan needs to be re-created
                # if scan is not present, its the last one, or force overwrite
                if (spec_scan.nr not in self.scan_dict.keys()) or \
                        (spec_scan.nr >= last_scan_number) or \
                        self.force_overwrite:
                    # rename init_motor_pos keys without prefix
                    init_motor_pos = {}
                    for key, value in spec_scan.init_motor_pos.items():
                        init_motor_pos[key.replace('INIT_MOPO_', '')] = value
                    # catching PR for itime in xu SpecScan missing
                    try:
                        int_time = float(spec_scan.itime)
                    except AttributeError:
                        int_time = 0.0
                    # create scan object
                    scan = Scan(int(spec_scan.nr),
                                cmd=spec_scan.command,
                                date=spec_scan.date,
                                time=spec_scan.time,
                                int_time=int_time,
                                header=spec_scan.header,
                                init_mopo=init_motor_pos)
                    self.scan_dict[spec_scan.nr] = scan
                    # check if the data needs to be read as well
                    if self.read_all_data:
                        self.read_scan_data(self.scan_dict[spec_scan.nr])

    def read_raw_scan_data(self, scan):
        """read_raw_scan_data

        Reads the data for a given scan object from raw source.

        Args:
            scan (Scan): scan object.

        """
        self.log.info('read_raw_scan_data for scan #{:d}'.format(scan.number))

        spec_scan = self.spec_file.__getattr__('scan{:d}'.format(scan.number))
        spec_scan.ReadData()
        scan.data = spec_scan.data
        spec_scan.ClearData()
        scan.meta['header'] = spec_scan.header



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2021 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


from numpy.core.records import fromarrays
import os
import h5py

from .source import Source
from .scan import Scan


class PalH5(Source):
    """PalH5

    Source implementation for PalH5 folder/files.

    Args:
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.

    Keyword Args:
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        read_all_data (bool): read all data on parsing.
          If false, data will be read only on demand.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    Attributes:
        log (logging.logger): logger instance from logging.
        name (str): name of the source
        scan_dict (dict(scan)): dict of scan objects with
          key being the scan number.
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        h5_file_exists(bool): if h5 file exists.
        read_all_data (bool): read all data on parsing.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    """
    def __init__(self, name, file_name, file_path, **kwargs):
        super().__init__(file_name, file_path, **kwargs)
        self.name = name

    def parse_raw(self):
        """parse_raw

        Parse the PalH5 folder and populate the `scan_dict`.

        """
        self.log.info('parse_raw')

        if not os.path.exists(self.file_path):
            self.log.error('File path does not exist!')
            return

        for root, subdirectories, files in os.walk(self.file_path):
            for sub_dir in sorted(subdirectories):
                # check for scan number in given range
                try:
                    scan_number = int(sub_dir)
                except ValueError:
                    self.log.exception('{:s} is no scan folder - skipping'.format(sub_dir))
                    continue

                if (scan_number >= self.start_scan_number) and \
                        ((scan_number <= self.stop_scan_number) or
                            (self.stop_scan_number == -1)):
                    last_scan_number = self.get_last_scan_number()
                    # check if Scan needs to be re-created
                    # if scan is not present, its the last one, or force overwrite
                    if (scan_number not in self.scan_dict.keys()) or \
                            (scan_number >= last_scan_number) or \
                            self.force_overwrite:
                        # create scan object

                        h5_file = os.path.join(self.file_path,
                                               self.file_name.format(scan_number),
                                               self.file_name.format(scan_number) + '.h5')

                        try:
                            with h5py.File(h5_file, 'r') as h5:
                                header = h5['R{0:04d}/header'.format(scan_number)]

                                init_motor_pos = {}
                                for key in header['motor_init_pos'].keys():
                                    init_motor_pos[key] = \
                                        header['motor_init_pos/{:s}'.format(key)][()]

                                # create scan object
                                try:
                                    # this is a fixQ fix
                                    int_time = float(header['scan_cmd'].asstr()[()].split(' ')[-1])
                                except ValueError:
                                    int_time = float(header['scan_cmd'].asstr()[()].split(' ')[-2])
                                scan = Scan(int(scan_number),
                                            cmd=header['scan_cmd'].asstr()[()],
                                            date=header['time'].asstr()[()].split(' ')[0],
                                            time=header['time'].asstr()[()].split(' ')[1],
                                            int_time=int_time,
                                            header='',
                                            init_mopo=init_motor_pos)
                                self.scan_dict[scan_number] = scan
                            # check if the data needs to be read as well
                            if self.read_all_data:
                                self.read_scan_data(self.scan_dict[scan_number])
                        except OSError:
                            self.log.warning('Could not open file {:s}'.format(h5_file))
                            continue

    def read_raw_scan_data(self, scan):
        """read_raw_scan_data

        Reads the data for a given scan object from Sardana NeXus file.

        Args:
            scan (Scan): scan object.

        """
        self.log.info('read_raw_scan_data for scan #{:d}'.format(scan.number))
        # try to open the file
        h5_file = os.path.join(self.file_path,
                               self.file_name.format(scan.number),
                               self.file_name.format(scan.number) + '.h5')

        with h5py.File(h5_file, 'r') as h5:
            entry = h5['R{0:04d}'.format(scan.number)]
            # iterate through data fields
            data_list = []
            dtype_list = []
            for key in entry['scan_dat'].keys():
                if '_raw' not in key:
                    data_list.append(entry['scan_dat'][key])
                    dtype_list.append((key,
                                      entry['scan_dat'][key].dtype,
                                      entry['scan_dat'][key].shape))
            if len(data_list) > 0:
                scan.data = fromarrays(data_list, dtype=dtype_list)
            else:
                scan.data = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest


@pytest.mark.parametrize('mysource, sname, scan_num, scan_delay',
                         [
                          ('source_spec', 'example_file_spec.spec', 1, -0.998557475),
                          ('source_pal', 'pal_file', 40, -33)
                         ])
def test_source(mysource, sname, scan_num, scan_delay, request):

    s = request.getfixturevalue(mysource)
    assert s.name == sname
    assert s.get_scan(scan_num).meta['number'] == scan_num
    assert s.get_scan(scan_num).data['delay'][0] == pytest.approx(scan_delay)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2021 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

from .. import config
import logging

from .scan import Scan

import os.path as path
from numpy.core.records import fromarrays
import nexusformat.nexus as nxs


class Source(object):
    """Source

    Class of default source implementation.

    Args:
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.

    Keyword Args:
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        nexus_file_name (str): name for generated nexus file.
        nexus_file_name_postfix (str): postfix for nexus file name.
        nexus_file_path (str): path for generated nexus file.
        read_all_data (bool): read all data on parsing.
          If false, data will be read only on demand.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_nexus (bool): use nexus file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of nexus file.

    Attributes:
        log (logging.logger): logger instance from logging.
        name (str): name of the source
        scan_dict (dict(scan)): dict of scan objects with
          key being the scan number.
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.
        nexus_file_name (str): name for generated nexus file.
        nexus_file_name_postfix (str): postfix for nexus file name.
        nexus_file_path (str): path for generated nexus file.
        nexus_file_exists(bool): if nexus file exists.
        read_all_data (bool): read all data on parsing.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_nexus (bool): use nexus file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of nexus file.

    """
    def __init__(self, file_name, file_path, **kwargs):
        self.log = logging.getLogger(__name__)
        self.log.setLevel(config.LOG_LEVEL)
        self.name = file_name
        self.scan_dict = {}
        self._start_scan_number = 0
        self._stop_scan_number = -1
        self.start_scan_number = kwargs.get('start_scan_number', 0)
        self.stop_scan_number = kwargs.get('stop_scan_number', -1)
        self.file_name = file_name
        self.file_path = file_path
        self.nexus_file_name_postfix = kwargs.get('nexus_file_name_postfix',
                                                  '.pyevaldata')
        self.nexus_file_name = kwargs.get('nexus_file_name', self.file_name)
        self.nexus_file_path = kwargs.get('nexus_file_path', self.file_path)
        self.check_nexus_file_exists()
        self.read_all_data = kwargs.get('read_all_data', False)
        self.read_and_forget = kwargs.get('read_and_forget', False)
        self.update_before_read = kwargs.get('update_before_read', False)
        self.use_nexus = kwargs.get('use_nexus', True)
        self.force_overwrite = kwargs.get('force_overwrite', False)

        # update from the source
        self.update()

    def __getattr__(self, attr):
        """__getattr__

        Allows to access scans as source attributes.

        Returns:
            scan (Scan): scan object.

        """
        if attr.startswith("scan"):
            index = attr[4:]

            try:
                scan_number = int(index)
            except ValueError:
                raise ValueError('Scan number must be convertable to an integer!')

            return self.get_scan(scan_number)
        else:
            raise AttributeError('\'{:s}\' has no attribute \'{:s}\''.format(__name__, attr))

    def __len__(self):
        """Returns length of ``scan_dict``"""
        return self.scan_dict.__len__()

    def update(self, scan_number_list=[]):
        """update

        update the ``scan_dict`` either from the raw source file/folder
        or from the nexus file.
        The optional ``scan_number_list`` runs the update only if required
        for the included scan.

        Attributes:
            scan_number_list (list[int]): explicit list of scans

        """

        if ~isinstance(scan_number_list, list):
            scan_number_list = list(scan_number_list)

        last_scan_number = self.get_last_scan_number()
        if (len(scan_number_list) == 0) \
                or (last_scan_number in scan_number_list) \
                or any(list(set(scan_number_list) - set(self.scan_dict.keys()))):

            self.log.info('Update source')

            if self.use_nexus:
                self.log.debug('Updating from nexus')
                # do not combine cases for better flow control
                if not self.nexus_file_exists:
                    self.log.debug('nexus file does not exist')
                    self.parse_raw()
                    self.save_all_scans_to_nexus()
                elif self.update_before_read:
                    self.log.debug('Update before read')
                    self.parse_raw()
                    self.save_all_scans_to_nexus()
                elif self.force_overwrite:
                    self.log.debug('Force overwrite')
                    self.parse_raw()
                    self.save_all_scans_to_nexus()
                else:
                    self.parse_nexus()
            else:
                self.log.debug('Updating from raw source')
                self.parse_raw()
        else:
            self.log.debug('Skipping update for scans {:s} '
                           'which are already present in '
                           'scan_dict.'.format(str(scan_number_list)))

    def parse_raw(self):
        """parse_raw

        Parse the raw source file/folder and populate the `scan_dict`.

        """
        raise NotImplementedError('Needs to be implemented!')

    def parse_nexus(self):
        """parse_nexus

        Parse the nexus file and populate the `scan_dict`.

        """
        self.log.info('parse_nexus')
        nxs_file_path = path.join(self.nexus_file_path, self.nexus_file_name)
        try:
            nxs_file = nxs.nxload(nxs_file_path, mode='r')
        except nxs.NeXusError:
            raise nxs.NeXusError('NeXus file \'{:s}\' does not exist!'.format(nxs_file_path))

        with nxs_file.nxfile:
            for entry in nxs_file:
                # check for scan number in given range
                if (nxs_file[entry].number >= self.start_scan_number) and \
                        ((nxs_file[entry].number <= self.stop_scan_number) or
                            (self.stop_scan_number == -1)):
                    last_scan_number = self.get_last_scan_number()
                    # check if Scan needs to be re-created
                    # if scan is not present, its the last one, or force overwrite
                    if (nxs_file[entry].number not in self.scan_dict.keys()) or \
                            (nxs_file[entry].number >= last_scan_number) or \
                            self.force_overwrite:
                        # create scan object
                        init_mopo = {}
                        for field in nxs_file[entry].init_mopo:
                            init_mopo[field] = nxs_file[entry]['init_mopo'][field]

                        scan = Scan(int(nxs_file[entry].number),
                                    cmd=nxs_file[entry].cmd,
                                    date=nxs_file[entry].date,
                                    time=nxs_file[entry].time,
                                    int_time=float(nxs_file[entry].int_time),
                                    header=nxs_file[entry].header,
                                    init_mopo=init_mopo)
                        self.scan_dict[nxs_file[entry].number] = scan
                        # check if the data needs to be read as well
                        if self.read_all_data:
                            self.read_scan_data(self.scan_dict[nxs_file[entry].number])

    def check_nexus_file_exists(self):
        """check_nexus_file_exists

        Check if the nexus file is present and set `self.nexus_file_exists`.

        """
        if path.exists(path.join(self.nexus_file_path, self.nexus_file_name)):
            self.nexus_file_exists = True
        else:
            self.nexus_file_exists = False

    def get_last_scan_number(self):
        """get_last_scan_number

        Return the number of the last scan in the `scan_dict`.
        If the `scan_dict` is empty return 0.

        """
        try:
            return sorted(self.scan_dict.keys())[-1]
        except IndexError:
            return 0

    def get_all_scan_numbers(self):
        """get_all_scan_numbers

        Return the all scan number from the `scan_dict`.

        """
        try:
            return sorted(self.scan_dict.keys())
        except IndexError:
            return 0

    def get_scan(self, scan_number, read_data=True, dismiss_update=False):
        """get_scan

        Returns a scan object from the scan dict determined by the scan_number.

        Args:
            scan_number (uint): number of the scan.
            read_data (bool, optional): read data from source.
              Defaults to `False`.
            dismiss_update (bool, optional): Dismiss update even if set as
              object attribute. Defaults to `False`.

        Returns:
            scan (Scan): scan object.

        """
        self.log.debug('get_scan')
        if self.update_before_read and not dismiss_update:
            self.update(scan_number)

        try:
            scan = self.scan_dict[scan_number]
        except KeyError:
            raise KeyError('Scan #{:d} not found in scan dict.'.format(scan_number))
        if read_data:
            self.read_scan_data(scan)
        return scan

    def get_scan_list(self, scan_number_list, read_data=True):
        """get_scan_list

        Returns a list of scan object from the `scan_dict` determined by
        the list of scan_number.

        Args:
            scan_number_list (list(uint)): list of numbers of the scan.
            read_data (bool, optional): read data from source.
              Defaults to `False`.

        Returns:
            scans (list(Scan)): list of scan object.

        """
        self.log.debug('get_scan_list')

        if self.update_before_read:
            self.update(scan_number_list)

        scans = []
        for scan_number in scan_number_list:
            scan = self.get_scan(scan_number, read_data, dismiss_update=True)

            scans.append(scan)

        return scans

    def get_scan_data(self, scan_number):
        """get_scan_data

        Returns data and meta information from a scan object from the `scan_dict`
        determined by the scan_number.

        Args:
            scan_number (uint): number of the scan.

        Returns:
            data (numpy.recarray[float]): scan data.
            meta (dict()): scan meta information.

        """
        self.log.debug('get_scan_data')

        scan = self.get_scan(scan_number)
        if scan.data is not None:
            data = scan.data.copy()
        else:
            data = None
        meta = scan.meta.copy()
        if self.read_and_forget:
            scan.clear_data()
        return data, meta

    def get_scan_list_data(self, scan_number_list):
        """get_scan_list_data

        Returns data and meta information for a list of scan objects from
        the `scan_dict` determined by the scan_numbers.

        Args:
            scan_number_list (list(uint)): list of numbers of the scan.

        Returns:
            data (list(numpy.recarray[float])): list of scan data.
            meta (list(dict())): list scan meta information.

        """
        self.log.debug('get_scan_list_data')

        data_list = []
        meta_list = []
        for scan in self.get_scan_list(scan_number_list):
            data_list.append(scan.data.copy())
            meta_list.append(scan.meta.copy())
            if self.read_and_forget:
                scan.clear_data()
        return data_list, meta_list

    def read_scan_data(self, scan):
        """read_scan_data

        Reads the data for a given scan object.

        Args:
            scan (Scan): scan object.

        """
        self.log.debug('read_scan_data for scan #{:d}'.format(scan.number))

        last_scan_number = self.get_last_scan_number()

        if (scan.data is None) or \
                (scan.number >= last_scan_number) or self.force_overwrite:
            if self.use_nexus:
                self.read_nexus_scan_data(scan)
            else:
                self.read_raw_scan_data(scan)
        else:
            self.log.debug('data not updated for scan #{:d}'.format(scan.number))

    def read_raw_scan_data(self, scan):
        """read_raw_scan_data

        Reads the data for a given scan object from raw source.

        Args:
            scan (Scan): scan object.

        """
        raise NotImplementedError('Needs to be implemented!')

    def read_nexus_scan_data(self, scan):
        """read_nexus_scan_data

        Reads the data for a given scan object from the nexus file.

        Args:
            scan (Scan): scan object.

        """
        self.log.debug('read_nexus_scan_data for scan #{:d}'.format(scan.number))
        # try to open the file
        nxs_file_path = path.join(self.nexus_file_path, self.nexus_file_name)
        try:
            nxs_file = nxs.nxload(nxs_file_path, mode='r')
        except nxs.NeXusError:
            raise nxs.NeXusError('NeXus file \'{:s}\' does not exist!'.format(nxs_file_path))

        entry_name = 'entry{:d}'.format(scan.number)
        # try to enter entry
        try:
            entry = nxs_file[entry_name]
        except nxs.NeXusError:
            self.log.exception('Entry #{:d} not present in NeXus file!'.format(scan.number))
            return
        # iterate through data fields
        data_list = []
        dtype_list = []
        for field in entry.data:
            data_list.append(entry.data[field])
            dtype_list.append((field, entry.data[field].dtype, entry.data[field].shape))
        if len(data_list) > 0:
            scan.data = fromarrays(data_list, dtype=dtype_list)
        else:
            scan.data = None

    def clear_scan_data(self, scan):
        """clear_scan_data

        Clear the data for a given scan object.

        Args:
            scan (Scan): scan object.

        """
        self.log.debug('clear_scan_data')

        scan.clear_data()

    def read_all_scan_data(self):
        """read_all_scan_data

        Reads the data for all scan objects in the `scan_dict` from source.

        """
        self.log.debug('read_all_scan_data')

        for scan_number, scan in self.scan_dict.items():
            self.read_scan_data(scan)

    def clear_all_scan_data(self):
        """clear_all_scan_data

        Clears the data for all scan objects in the `scan_dict`.

        """
        self.log.debug('clear_all_scan_data')

        for scan_number, scan in self.scan_dict.items():
            self.clear_scan_data(scan)

    def save_scan_to_nexus(self, scan, nxs_file=''):
        """save_scan_to_nexus

        Saves a scan to the nexus file.

        """
        if nxs_file == '':
            nxs_file = self.get_nexus_file()

        entry_name = 'entry{:d}'.format(scan.number)

        # evaluate if we need to forget the data again
        if scan.data is None:
            clear_data = True
        else:
            clear_data = False
        # read the raw data
        self.read_raw_scan_data(scan)

        self.log.info('save_scan_to_nexus for scan #{:d}'.format(scan.number))
        with nxs_file.nxfile:
            # if the entry already exists, it must be deleted in advance
            try:
                del nxs_file[entry_name]
            except nxs.NeXusError:
                pass
            # (re-)create entry
            entry = nxs_file[entry_name] = nxs.NXentry()
            # iterate meta information
            for key, value in scan.meta.items():
                if key == 'init_mopo':
                    # create dedicated collection for initial motor positions
                    entry['init_mopo'] = nxs.NXcollection()
                    # iterate through initial motor positions
                    for mopo_key, mopo_value in scan.meta['init_mopo'].items():
                        entry.init_mopo[mopo_key] = nxs.NXfield(mopo_value)
                else:
                    # add meta information as attribute to entry
                    entry.attrs[key] = value
            # create dedicated collection for data
            entry['data'] = nxs.NXcollection()
            # check if there is any data present at all
            if scan.data is not None:
                # iterate data
                for col in scan.data.dtype.names:
                    entry.data[col] = nxs.NXfield(scan.data[col])
                # clear data of the scan if it was not present before
                # or read and forget
                if clear_data or self.read_and_forget:
                    scan.clear_data()

    def save_all_scans_to_nexus(self):
        """save_all_scans_to_nexus

        Saves all scan objects in the `scan_dict` to the nexus file.

        """
        self.log.info('save_all_scans_to_nexus')
        nxs_file = self.get_nexus_file()
        try:
            last_scan_in_nexus = sorted(int(num.strip('entry')) for num in nxs_file.keys())[-1]
        except IndexError:
            last_scan_in_nexus = -1

        for scan_number, scan in self.scan_dict.items():
            entry_name = 'entry{:d}'.format(scan.number)
            try:
                _ = nxs_file[entry_name]
                scan_in_nexus = True
            except KeyError:
                scan_in_nexus = False

            if (not scan_in_nexus) or (scan.number >= last_scan_in_nexus) \
                    or self.force_overwrite:
                self.save_scan_to_nexus(scan, nxs_file)

    def get_nexus_file(self, mode='rw'):
        """get_nexus_file

        Return the file handle to the NeXus file in a given ``mode```.

        Args:
            mode (str, optional): file mode. defaults to 'rw'.

        Returns:
            nxs_file (NXFile): file handle to NeXus file.

        """
        self.log.debug('get_nexus_file')
        try:
            nxs_file = nxs.nxload(path.join(self.nexus_file_path, self.nexus_file_name), mode='rw')
        except nxs.NeXusError:
            nxs.NXroot().save(path.join(self.nexus_file_path, self.nexus_file_name))
            nxs_file = nxs.nxload(path.join(self.nexus_file_path, self.nexus_file_name), mode='rw')
        return nxs_file

    @property
    def nexus_file_name(self):
        return self._nexus_file_name

    @nexus_file_name.setter
    def nexus_file_name(self, nexus_file_name):
        self._nexus_file_name = nexus_file_name + self.nexus_file_name_postfix + '.nxs'

    @property
    def start_scan_number(self):
        return self._start_scan_number

    @start_scan_number.setter
    def start_scan_number(self, start_scan_number):
        if start_scan_number < 0:
            self.log.warning('start_scan_number must not be negative!')
            return
        elif (start_scan_number > self.stop_scan_number) and (self.stop_scan_number > -1):
            self.log.warning('start_scan_number must be <= stop_scan_number!')
            return
        else:
            self._start_scan_number = start_scan_number

    @property
    def stop_scan_number(self):
        return self._stop_scan_number

    @stop_scan_number.setter
    def stop_scan_number(self, stop_scan_number):
        if stop_scan_number < -1:
            self.log.warning('stop_scan_number cannot be smaller than -1!')
            return
        elif (stop_scan_number < self.start_scan_number) and (stop_scan_number > -1):
            self.log.warning('stop_scan_number must be >= start_scan_number!')
            return
        else:
            self._stop_scan_number = stop_scan_number



import logging
import sys
from . import config

logging.basicConfig(stream=sys.stdout,
                    format=config.LOG_FORMAT)

from .evaluation import Evaluation
from . import io

__all__ = ['Evaluation', 'io']

__version__ = '1.3.7'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
from scipy.stats import binned_statistic


def edges4grid(grid):
    """Creates a vector of the corresponding edges for a grid vector. """
    binwidth = np.diff(grid)
    edges = np.hstack([grid[0]-binwidth[0]/2, grid[0:-1] +
                       binwidth/2, grid[-1]+binwidth[-1]/2])

    return edges, binwidth


def bin_data(y, x, X, statistic='mean'):
    """Bin data y(x) on new grid X using a statistic type. """

    # get only unmasked data
    idx = ~np.ma.getmask(x)
    idy = ~np.ma.getmask(y)
    y = y[idx & idy].flatten('F')
    x = x[idx & idy].flatten('F')
    idX = ~np.ma.getmask(X)
    X = np.unique(np.sort(X[idX].flatten('F')))
    # set non-finite values to 0
    y[~np.isfinite(y)] = 0
    # create bins for the grid
    edges, _ = edges4grid(X)

    if np.array_equal(x, X):
        # no binning since the new grid is the same as the old one
        Y = y
        bins = np.ones_like(Y)
        n = np.ones_like(Y)
    else:
        # do the binning and get the Y results
        Y, _, bins = binned_statistic(x, y, statistic, edges)
        bins = bins.astype(np.int_)

        n = np.bincount(bins[bins > 0], minlength=len(X)+1)
        n = n[1:len(X)+1]

    if np.array_equal(x, X) and statistic != 'sum':
        Ystd = np.zeros_like(Y)
        Xstd = np.zeros_like(X)
        Yerr = np.zeros_like(Y)
        Xerr = np.zeros_like(X)
    else:
        # calculate the std of X and Y
        if statistic == 'sum':
            Ystd = np.sqrt(Y)
            Yerr = Ystd
        else:
            Ystd, _, _ = binned_statistic(x, y, 'std', edges)
            Yerr = Ystd/np.sqrt(n)

        Xstd, _, _ = binned_statistic(x, x, 'std', edges)
        Xerr = Xstd/np.sqrt(n)

    # remove NaNs
    Y = Y[n > 0]
    X = X[n > 0]
    Yerr = Yerr[n > 0]
    Xerr = Xerr[n > 0]
    Ystd = Ystd[n > 0]
    Xstd = Xstd[n > 0]

    return Y, X, Yerr, Xerr, Ystd, Xstd, edges, bins, n



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pytest


def test_evaluation(evaluation):
    data = evaluation
    assert data.cdef == {}
    assert data.clist == []
    cdef = {
        'Pumped_demod':  'Pumped-(Unpumped-mean(Unpumped))',
        'PumpedM_demod': 'PumpedM-(UnpumpedM-mean(UnpumpedM))',
        'M':             'Pumped_demod-PumpedM_demod',
    }
    data.cdef = cdef
    data.xcol = 'delay'
    data.clist = ['M']
    data1 = data.get_scan_data(1)
    assert data1['delay'][0] == pytest.approx(-0.998557475)
    y, x, yerr, xerr, name = data.plot_scans([1])
    assert y[data.clist[0]][0] == pytest.approx(0.02183873769)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

import numpy as np
import collections
import matplotlib.pyplot as plt
import matplotlib as mpl
import re
from uncertainties import unumpy
from .helpers import bin_data


class Evaluation(object):
    """Evaluation

    Attributes:
        clist (List[str])       : List of counter names to evaluate.
        cdef (Dict{str:str})    : Dict of predefined counter names and
                                  definitions.
        xcol (str)              : spec counter or motor to plot as x-axis.
        t0 (float)              : approx. time zero for delay scans to
                                  determine the unpumped region of the data
                                  for normalization.
        custom_counters (List[str]): List of custom counters - default is []
        math_keys (List[str])    : List of keywords which are not replaced in
                                  counter names
        statistic_type  (str)    : 'gauss' for normal averaging,
                                  'poisson' for counting statistics
        propagate_errors  (bool) : whether to propagate errors or not

    """

    def __init__(self, source):
        # properties
        self.source = source
        self.clist = []
        self.cdef = {}
        self.xcol = ''
        self.t0 = 0
        self.custom_counters = []
        self.math_keys = ['mean', 'sum', 'diff', 'max', 'min', 'round', 'abs',
                          'sin', 'cos', 'tan', 'arcsin', 'arccos', 'arctan',
                          'pi', 'exp', 'log', 'log10', 'sqrt']
        self.statistic_type = 'gauss'
        self.propagate_errors = True
        self.apply_data_filter = False
        self.data_filters = ['evaluatable statement']

    def get_clist(self):
        """Return the list of counters to evaluate as list even if they are
        provided as Dict by the user.
        This method is only for backward compatibility to older versions.

        Returns:
            clist (List[str]): List of counter names to evaluate.

        """

        if isinstance(self.clist, dict):
            # the clist property is a dict, so retrun its keys as list
            clist = list(self.clist.keys())
        else:
            clist = list(self.clist)

        return clist

    def get_last_fig_number(self):
        """Return the last figure number of all opened figures for plotting
        data in the same figure during for-loops.

        Returns:
            fig_number (int): Last figure number of all opened figures.

        """

        try:
            # get the number of all opened figures
            fig_number = mpl._pylab_helpers.Gcf.get_active().num
        except Exception:
            # there are no figures open
            fig_number = 1

        return fig_number

    def get_next_fig_number(self):
        """Return the number of the next figure for plotting data in the
        same figure during for-loops.

        Returns:
            next_fig_number (int): Next figure number of all opened figures.

        """

        return self.get_last_fig_number() + 1

    def traverse_counters(self, clist, spec_cols=''):
        """Traverse all counters and replace all predefined counter definitions.
        Returns also the included spec base counters for error propagation.

        Args:
            clist    (list) : Initial counter list.
            spec_cols (list) : Counters in spec file.

        Returns:
            resolved_counters (list): Resolved counters.
            base_counters (list)    : Base counters.

        """

        resolved_counters = []
        base_counters = []

        for col_name in clist:
            col_string, res_base_counters = self.resolve_counter_name(col_name, spec_cols)

            resolved_counters.append(col_string)
            base_counters.extend(res_base_counters)

        # remove duplicates using list(set())
        return resolved_counters, list(set(base_counters))

    def resolve_counter_name(self, col_name, spec_cols=''):
        """Replace all predefined counter definitions in a counter name.
        The function works recursively.

        Args:
            col_name (str) : Initial counter string.

        Returns:l
            col_string (str): Resolved counter string.

        """

        recall = False  # boolean to stop recursive calls

        base_counters = []

        col_string = col_name

        for find_cdef in self.cdef.keys():
            # check for all predefined counters
            search_pattern = r'\b' + find_cdef + r'\b'
            if re.search(search_pattern, col_string) is not None:
                if self.cdef[find_cdef] in spec_cols:
                    # this counter definition is a base spec counter
                    base_counters.append(self.cdef[find_cdef])
                # found a predefined counter
                # recursive call if predefined counter must be resolved again
                recall = True
                # replace the counter definition in the string
                (col_string, _) = re.subn(search_pattern,
                                          '(' + self.cdef[find_cdef] + ')', col_string)

        if recall:
            # do the recursive call
            col_string, rec_base_counters = self.resolve_counter_name(col_string, spec_cols)
            base_counters.extend(rec_base_counters)

        for find_cdef in spec_cols:
            # check for all base spec counters
            search_pattern = r'\b' + find_cdef + r'\b'
            if re.search(search_pattern, col_string) is not None:
                base_counters.append(find_cdef)

        return col_string, base_counters

    def col_string_to_eval_string(self, col_string, array_name='spec_data'):
        """Use regular expressions in order to generate an evaluateable string
        from the counter string in order to append the new counter to the
        spec data.

        Args:
            col_string (str) : Definition of the counter.
            mode (int)      : Flag for different modes

        Returns:
            eval_string (str): Evaluateable string to add the new counter
                              to the spec data.

        """

        # search for alphanumeric counter names in col_string
        iterator = re.finditer(
            '([0-9]*[a-zA-Z\_]+[0-9]*[a-zA-Z]*)*', col_string)
        # these are keys which should not be replaced but evaluated
        math_keys = list(self.math_keys)
        keys = math_keys.copy()

        for key in iterator:
            # traverse all found counter names
            if len(key.group()) > 0:
                # the match is > 0
                if not key.group() in keys:
                    # the counter name is not in the keys list

                    # remember this counter name in the key list in order
                    # not to replace it again
                    keys.append(key.group())
                    # the actual replacement
                    (col_string, _) = re.subn(r'\b'+key.group()+r'\b',
                                              array_name + '[\'' + key.group() + '\']', col_string)

        # add 'np.' prefix to numpy functions/math keys
        for mk in math_keys:
            if mk != '0x0001FFFF':
                (col_string, _) = re.subn(r'\b' + mk + r'\b', 'np.' + mk, col_string)
        return col_string

    def add_custom_counters(self, spec_data, scan_num, base_counters):
        """Add custom counters to the spec data array.
        This is a stub for child classes.

        Args:
            spec_data (ndarray) : Data array from the spec scan.
            scan_num (int)  : Scan number of the spec scan.
            base_counters list(str) : List of the base spec and custom counters
                                     from the clist and xcol.

        Returns:
            spec_data (ndarray): Updated data array from the spec scan.

        """
        return spec_data

    def filter_data(self, data):
        """filter_data

        Args:
            data (TYPE): DESCRIPTION.

        Returns:
            TYPE: DESCRIPTION.

        """
        res = []
        for data_filter in self.data_filters:
            name, _ = self.resolve_counter_name(data_filter)
            idx = eval(self.col_string_to_eval_string(name, array_name='data'))
            if len(res) == 0:
                res = idx
            else:
                res = np.logical_and(res, idx)

        data_list = []
        dtype_list = []
        for name in data.dtype.names:
            data_list.append(data[name][res])
            dtype_list.append((name,
                               data[name][res].dtype,
                               data[name][res].shape))
        return np.core.records.fromarrays(data_list, dtype=dtype_list)

    def get_scan_data(self, scan_num):
        """

        Args:
            scan_num (TYPE): DESCRIPTION.

        Returns:
            TYPE: DESCRIPTION.

        """
        data, meta = self.source.get_scan_data(scan_num)
        if self.apply_data_filter:
            data = self.filter_data(data)
        return data

    def get_scan_list_data(self, scan_list):
        """

        Args:
            scan_num (TYPE): DESCRIPTION.

        Returns:
            TYPE: DESCRIPTION.

        """
        data_list, meta_list = self.source.get_scan_list_data(scan_list)
        if self.apply_data_filter:
            for i, data in enumerate(data_list):
                data_list[i] = self.filter_data(data)
        return data_list

    def avg_N_bin_scans(self, scan_list, xgrid=np.array([]), binning=True):
        """Averages data defined by the counter list, clist, onto an optional
        xgrid. If no xgrid is given the x-axis data of the first scan in the
        list is used instead.

        Args:
            scan_list (List[int])      : List of scan numbers.
            xgrid (Optional[ndarray]) : Grid to bin the data to -
                                        default in empty so use the
                                        x-axis of the first scan.

        Returns:
            avg_data (ndarray) : Averaged data for the scan list.
            std_data (ndarray) : Standart derivation of the data for the scan list.
            err_data (ndarray) : Error of the data for the scan list.
            name (str)        : Name of the data set.

        """

        # generate the name of the data set from the spec file name and scan_list
        name = self.source.name + " #{0:04d}".format(scan_list[0])

        # get the counters which should be evaluated
        clist = self.get_clist()
        if not clist:
            raise Exception('No clist is defined. Do not know what to plot!')
            return
        # process also the xcol as counter in order to allow for newly defined xcols
        if not self.xcol:
            raise Exception('No xcol is defined. Do not know what to plot!')
            return
        if self.xcol not in clist:
            clist.append(self.xcol)

        spec_cols = []
        concat_data = np.array([])

        data_list = self.get_scan_list_data(scan_list)

        for i, (spec_data, scan_num) in enumerate(zip(data_list, scan_list)):
            # traverse the scan list and read data
            # try:
            #     # try to read the motors and data of this scan
            #     spec_data = self.get_scan_data(scan_num)
            # except Exception:
            #     raise
            #     print('Scan #' + scan_num + ' not found, skipping')

            if i == 0 or len(spec_cols) == 0:  # we need to evaluate this only once
                # these are the base spec counters which are present in the data
                # file plus custom counters
                spec_cols = list(
                    set(list(spec_data.dtype.names) + self.custom_counters))

                # resolve the clist and retrieve the resolves counters and the
                # necessary base spec counters for error propagation
                resolved_counters, base_counters = self.traverse_counters(
                    clist, spec_cols)

                # counter names and resolved strings for further calculations
                if self.statistic_type == 'poisson' or self.propagate_errors:
                    # for error propagation we just need the base spec counters
                    # and the xcol
                    col_names = base_counters[:]
                    col_strings = base_counters[:]
                    # add the xcol to both lists
                    col_names.append(self.xcol)
                    col_strings.append(resolved_counters[clist.index(self.xcol)])
                else:
                    # we need to average the resolved counters
                    col_names = clist[:]
                    col_strings = resolved_counters[:]

                # create the dtype of the return array
                dtypes = []
                for col_name in clist:
                    dtypes.append((col_name, '<f8'))

            # add custom counters if defined
            spec_data = self.add_custom_counters(spec_data, scan_num, base_counters)

            data = np.array([])
            # read data into data array
            for col_string, col_name in zip(col_strings, col_names):
                # traverse the counters in the clist and append to data if not
                # already present
                eval_string = self.col_string_to_eval_string(
                    col_string, array_name='spec_data')

                if len(data) == 0:
                    data = np.array(eval(eval_string), dtype=[(col_name, float)])
                elif col_name not in data.dtype.names:
                    data = eval('np.lib.recfunctions.append_fields(data,\''
                                + col_name + '\',data=(' + eval_string
                                + '), dtypes=float, asrecarray=True, usemask=True)')

            if i > 0:
                # this is not the first scan in the list so append the data to
                # the concatenated data array
                concat_data = np.concatenate((concat_data, data), axis=0)
            else:
                concat_data = data

                if len(xgrid) == 0:
                    # if no xgrid is given we use the xData of the first scan instead
                    xgrid = concat_data[self.xcol]

        # remove xcol from clist and resolved counters for further treatment
        del resolved_counters[clist.index(self.xcol)]
        clist.remove(self.xcol)

        try:
            # bin the concatenated data to the xgrid
            # if a custom counter was calculated it might have a different length
            # than the spec counters which will result in an error while binning data
            # from a default spec counter and a custom counter.
            if binning:
                xgrid_reduced, _, _, _, _, _, _, _, _ = bin_data(
                    concat_data[self.xcol], concat_data[self.xcol], xgrid)
            else:
                xgrid_reduced = xgrid
            # create empty arrays for averages, std and errors
            avg_data = np.recarray(np.shape(xgrid_reduced)[0], dtype=dtypes)
            std_data = np.recarray(np.shape(xgrid_reduced)[0], dtype=dtypes)
            err_data = np.recarray(np.shape(xgrid_reduced)[0], dtype=dtypes)

            if self.statistic_type == 'poisson':
                bin_stat = 'sum'
            else:  # gauss
                bin_stat = 'mean'

            if binning:
                if self.statistic_type == 'poisson' or self.propagate_errors:
                    # propagate errors using the uncertainties package

                    # create empty dict for uncertainties data arrays
                    unc_data_err = {}
                    unc_data_std = {}

                    for col in base_counters:
                        # for all cols in the clist bin the data to the xgrid an calculate
                        # the averages, stds and errors
                        y, avg_data[self.xcol], yerr, err_data[self.xcol], ystd, \
                            std_data[self.xcol], _, _, _ = bin_data(concat_data[col],
                                                                    concat_data[self.xcol],
                                                                    xgrid_reduced,
                                                                    statistic=bin_stat)
                        # add spec base counters to uncData arrays
                        unc_data_std[col] = unumpy.uarray(y, ystd)
                        unc_data_err[col] = unumpy.uarray(y, yerr)

                    for col_name, col_string in zip(clist, resolved_counters):
                        eval_string = self.col_string_to_eval_string(
                            col_string, array_name='unc_data_err')
                        temp = eval(eval_string)

                        avg_data[col_name] = unumpy.nominal_values(temp)
                        err_data[col_name] = unumpy.std_devs(temp)

                        eval_string = self.col_string_to_eval_string(
                            col_string, array_name='unc_data_std')
                        temp = eval(eval_string)
                        std_data[col_name] = unumpy.std_devs(temp)
                else:
                    # no error propagation but averaging of individual scans
                    for col in clist:
                        # for all cols in the clist bin the data to the xgrid an calculate
                        # the averages, stds and errors
                        avg_data[col], avg_data[self.xcol], err_data[col], \
                            err_data[self.xcol], std_data[col], std_data[self.xcol], _, _, \
                            _ = bin_data(concat_data[col],
                                         concat_data[self.xcol],
                                         xgrid_reduced,
                                         statistic=bin_stat)
            else:
                for col_name, col_string in zip(clist, resolved_counters):
                    eval_string = self.col_string_to_eval_string(
                        col_string, array_name='spec_data')
                    temp = eval(eval_string)
                    avg_data[col_name] = temp
                    avg_data[self.xcol] = concat_data[self.xcol]
                    err_data[col_name] = 0
                    err_data[self.xcol] = 0
                    std_data[col_name] = 0
                    std_data[self.xcol] = 0

        except Exception:
            raise
            print('xcol and ycol must have the same length --> probably you try plotting a custom'
                  ' counter together with a spec counter')

        return avg_data, std_data, err_data, name

    def plot_scans(self, scan_list, ylims=[], xlims=[], fig_size=[], xgrid=[],
                   yerr='std', xerr='std', norm2one=False, binning=True,
                   label_text='', title_text='', skip_plot=False, grid_on=True,
                   ytext='', xtext='', fmt='-o'):
        """Plot a list of scans from the spec file.
        Various plot parameters are provided.
        The plotted data are returned.

        Args:
            scan_list (List[int])        : List of scan numbers.
            ylims (Optional[ndarray])   : ylim for the plot.
            xlims (Optional[ndarray])   : xlim for the plot.
            fig_size (Optional[ndarray]) : Figure size of the figure.
            xgrid (Optional[ndarray])   : Grid to bin the data to -
                                          default in empty so use the
                                          x-axis of the first scan.
            yerr (Optional[ndarray])    : Type of the errors in y: [err, std, none]
                                          default is 'std'.
            xerr (Optional[ndarray])    : Type of the errors in x: [err, std, none]
                                          default is 'std'.
            norm2one (Optional[bool])   : Norm transient data to 1 for t < t0
                                          default is False.
            label_text (Optional[str])   : Label of the plot - default is none.
            title_text (Optional[str])   : Title of the figure - default is none.
            skip_plot (Optional[bool])   : Skip plotting, just return data
                                          default is False.
            grid_on (Optional[bool])     : Add grid to plot - default is True.
            ytext (Optional[str])       : y-Label of the plot - defaults is none.
            xtext (Optional[str])       : x-Label of the plot - defaults is none.
            fmt (Optional[str])         : format string of the plot - defaults is -o.

        Returns:
            y2plot (OrderedDict)    : y-data which was plotted.
            x2plot (ndarray)        : x-data which was plotted.
            yerr2plot (OrderedDict) : y-error which was plotted.
            xerr2plot (ndarray)     : x-error which was plotted.
            name (str)              : Name of the data set.

        """

        # initialize the y-data as ordered dict in order to allow for multiple
        # counters at the same time
        y2plot = collections.OrderedDict()
        yerr2plot = collections.OrderedDict()

        # get the averaged data, stds and errors for the scan list and the xgrid
        avg_data, std_data, err_data, name = self.avg_N_bin_scans(
            scan_list, xgrid=xgrid, binning=binning)

        # set the error data
        if xerr == 'std':
            xerr_data = std_data
        elif xerr == 'err':
            xerr_data = err_data
        else:
            xerr_data = np.zeros_like(std_data)

        if yerr == 'std':
            yerr_data = std_data
        elif yerr == 'err':
            yerr_data = err_data
        else:
            yerr_data = np.zeros_like(std_data)

        # set x-data and errors
        x2plot = avg_data[self.xcol]
        xerr2plot = xerr_data[self.xcol]

        # plot all keys in the clist
        clist = self.get_clist()
        for col in clist:
            # traverse the counter list

            # save the counter data and errors in the ordered dictionary
            y2plot[col] = avg_data[col]
            yerr2plot[col] = yerr_data[col]

            if norm2one:
                # normalize the y-data to 1 for t < t0
                # just makes sense for delay scans
                before_zero = y2plot[col][x2plot <= self.t0]
                y2plot[col] = y2plot[col]/np.mean(before_zero)
                yerr2plot[col] = yerr2plot[col]/np.mean(before_zero)

            if len(label_text) == 0:
                # if no label_text is given use the counter name
                lt = col
            else:
                if len(clist) > 1:
                    # for multiple counters add the counter name to the label
                    lt = label_text + ' | ' + col
                else:
                    # for a single counter just use the label_text
                    lt = label_text

            if not skip_plot:
                # plot the errorbar for each counter
                if (xerr == 'none') & (yerr == 'none'):
                    plt.plot(x2plot, y2plot[col], fmt, label=lt)
                else:
                    plt.errorbar(
                        x2plot, y2plot[col], fmt=fmt, label=lt,
                        xerr=xerr2plot, yerr=yerr2plot[col])

        if not skip_plot:
            # add a legend, labels, title and set the limits and grid
            plt.legend(frameon=True, loc=0, numpoints=1)
            plt.xlabel(self.xcol)
            if xlims:
                plt.xlim(xlims)
            if ylims:
                plt.ylim(ylims)
            if len(title_text) > 0:
                plt.title(title_text)
            else:
                plt.title(name)
            if len(xtext) > 0:
                plt.xlabel(xtext)

            if len(ytext) > 0:
                plt.ylabel(ytext)

            if grid_on:
                plt.grid(True)

        return y2plot, x2plot, yerr2plot, xerr2plot, name

    def plot_mesh_scan(self, scan_num, skip_plot=False, grid_on=False, ytext='', xtext='',
                       levels=20, cbar=True):
        """Plot a single mesh scan from the spec file.
        Various plot parameters are provided.
        The plotted data are returned.

        Args:
            scan_num (int)               : Scan number of the spec scan.
            skip_plot (Optional[bool])   : Skip plotting, just return data
                                          default is False.
            grid_on (Optional[bool])     : Add grid to plot - default is False.
            ytext (Optional[str])       : y-Label of the plot - defaults is none.
            xtext (Optional[str])       : x-Label of the plot - defaults is none.
            levels (Optional[int])      : levels of contour plot - defaults is 20.
            cbar (Optional[bool])       : Add colorbar to plot - default is True.

        Returns:
            xx, yy, zz              : x,y,z data which was plotted

        """

        from matplotlib.mlab import griddata
        from matplotlib import gridspec

        # read data from spec file
        try:
            # try to read data of this scan
            spec_data = self.get_scan_data(scan_num)
        except Exception:
            print('Scan #' + int(scan_num) + ' not found, skipping')

        dt = spec_data.dtype
        dt = dt.descr

        xmotor = dt[0][0]
        ymotor = dt[1][0]

        X = spec_data[xmotor]
        Y = spec_data[ymotor]

        xx = np.sort(np.unique(X))
        yy = np.sort(np.unique(Y))

        clist = self.get_clist()

        if len(clist) > 1:
            print('WARNING: Only the first counter of the clist is plotted.')

        Z = spec_data[clist[0]]

        zz = griddata(X, Y, Z, xx, yy, interp='linear')

        if not skip_plot:

            if cbar:
                gs = gridspec.GridSpec(4, 2,
                                       width_ratios=[3, 1],
                                       height_ratios=[0.2, 0.1, 1, 3]
                                       )
                k = 4
            else:
                gs = gridspec.GridSpec(2, 2,
                                       width_ratios=[3, 1],
                                       height_ratios=[1, 3]
                                       )
                k = 0

            ax1 = plt.subplot(gs[0+k])

            plt.plot(xx, np.mean(zz, 0), label='mean')

            plt.plot(xx, zz[np.argmax(np.mean(zz, 1)), :], label='peak')

            plt.xlim([min(xx), max(xx)])
            plt.legend(loc=0)
            ax1.xaxis.tick_top()
            if grid_on:
                plt.grid(True)

            plt.subplot(gs[2+k])

            plt.contourf(xx, yy, zz, levels, cmap='viridis')

            plt.xlabel(xmotor)
            plt.ylabel(ymotor)

            if len(xtext) > 0:
                plt.xlabel(xtext)

            if len(ytext) > 0:
                plt.ylabel(ytext)

            if grid_on:
                plt.grid(True)

            if cbar:
                cb = plt.colorbar(cax=plt.subplot(
                    gs[0]), orientation='horizontal')
                cb.ax.xaxis.set_ticks_position('top')
                cb.ax.xaxis.set_label_position('top')

            ax4 = plt.subplot(gs[3+k])

            plt.plot(np.mean(zz, 1), yy)
            plt.plot(zz[:, np.argmax(np.mean(zz, 0))], yy)
            plt.ylim([np.min(yy), np.max(yy)])

            ax4.yaxis.tick_right()
            if grid_on:
                plt.grid(True)

        return xx, yy, zz

    def plot_scan_sequence(self, scan_sequence, ylims=[], xlims=[], fig_size=[],
                           xgrid=[], yerr='std', xerr='std', norm2one=False,
                           binning=True, sequence_type='', label_text='',
                           title_text='', skip_plot=False, grid_on=True, ytext='',
                           xtext='', fmt='-o'):
        """Plot a list of scans from the spec file.
        Various plot parameters are provided.
        The plotted data are returned.

        Args:
            scan_sequence (ndarray[List[int]
                          , int/str])   : Sequence of scan lists and parameters.
            ylims (Optional[ndarray])   : ylim for the plot.
            xlims (Optional[ndarray])   : xlim for the plot.
            fig_size (Optional[ndarray]) : Figure size of the figure.
            xgrid (Optional[ndarray])   : Grid to bin the data to -
                                          default in empty so use the
                                          x-axis of the first scan.
            yerr (Optional[ndarray])    : Type of the errors in y: [err, std, none]
                                          default is 'std'.
            xerr (Optional[ndarray])    : Type of the errors in x: [err, std, none]
                                          default is 'std'.
            norm2one (Optional[bool])   : Norm transient data to 1 for t < t0
                                          default is False.
            sequence_type (Optional[str]): Type of the sequence: [fluence, delay,
                                          energy, theta, position, voltage, none,
                                          text] - default is enumeration.
            label_text (Optional[str])   : Label of the plot - default is none.
            title_text (Optional[str])   : Title of the figure - default is none.
            skip_plot (Optional[bool])   : Skip plotting, just return data
                                          default is False.
            grid_on (Optional[bool])     : Add grid to plot - default is True.
            ytext (Optional[str])       : y-Label of the plot - defaults is none.
            xtext (Optional[str])       : x-Label of the plot - defaults is none.
            fmt (Optional[str])         : format string of the plot - defaults is -o.

        Returns:
            sequence_data (OrderedDict) : Dictionary of the averaged scan data.
            parameters (ndarray)       : Parameters of the sequence.
            names (List[str])          : List of names of each data set.
            label_texts (List[str])     : List of labels for each data set.

        """

        # initialize the return data
        sequence_data = collections.OrderedDict()
        names = []
        label_texts = []
        parameters = []

        for i, (scan_list, parameter) in enumerate(scan_sequence):
            # traverse the scan sequence

            parameters.append(parameter)
            # format the parameter as label text of this plot if no label text
            # is given
            if len(label_text) == 0:
                if sequence_type == 'fluence':
                    lt = str.format('{:.2f}  mJ/cm²', parameter)
                elif sequence_type == 'delay':
                    lt = str.format('{:.2f}  ps', parameter)
                elif sequence_type == 'energy':
                    lt = str.format('{:.2f}  eV', parameter)
                elif sequence_type == 'theta':
                    lt = str.format('{:.2f}  deg', parameter)
                elif sequence_type == 'temperature':
                    lt = str.format('{:.2f}  K', parameter)
                elif sequence_type == 'position':
                    lt = str.format('{:.2f}  mm', parameter)
                elif sequence_type == 'voltage':
                    lt = str.format('{:.2f}  V', parameter)
                elif sequence_type == 'current':
                    lt = str.format('{:.2f}  A', parameter)
                elif sequence_type == 'scans':
                    lt = str(scan_list)
                elif sequence_type == 'none':
                    # no parameter for single scans
                    lt = ''
                elif sequence_type == 'text':
                    # parameter is a string
                    lt = parameter
                else:
                    # no sequence type is given --> enumerate
                    lt = str.format('#{}', i+1)
            else:
                lt = label_text[i]

            # get the plot data for the scan list
            y2plot, x2plot, yerr2plot, xerr2plot, name = self.plot_scans(
                scan_list,
                ylims=ylims,
                xlims=xlims,
                fig_size=fig_size,
                xgrid=xgrid,
                yerr=yerr,
                xerr=xerr,
                norm2one=norm2one,
                binning=binning,
                label_text=lt,
                title_text=title_text,
                skip_plot=skip_plot,
                grid_on=grid_on,
                ytext=ytext,
                xtext=xtext,
                fmt=fmt
            )

            if self.xcol not in sequence_data.keys():
                # if the xcol is not in the return data dict - add the key
                sequence_data[self.xcol] = []
                sequence_data[self.xcol + 'Err'] = []

            # add the x-axis data to the return data dict
            sequence_data[self.xcol].append(x2plot)
            sequence_data[self.xcol + 'Err'].append(xerr2plot)

            for counter in y2plot:
                # traverse all counters in the data set
                if counter not in sequence_data.keys():
                    # if the counter is not in the return data dict - add the key
                    sequence_data[counter] = []
                    sequence_data[counter + 'Err'] = []

                # add the counter data to the return data dict
                sequence_data[counter].append(y2plot[counter])
                sequence_data[counter + 'Err'].append(yerr2plot[counter])

            # append names and labels to their lists
            names.append(name)
            label_texts.append(lt)

        return sequence_data, parameters, names, label_texts

    def export_scan_sequence(self, scan_sequence, path, fileName, yerr='std',
                             xerr='std', xgrid=[], norm2one=False, binning=True):
        """Exports spec data for each scan list in the sequence as individual file.

        Args:
            scan_sequence (ndarray[List[int]
                          , int/str])   : Sequence of scan lists and parameters.
            path (str)                  : Path of the file to export to.
            fileName (str)              : Name of the file to export to.
            yerr (Optional[ndarray])    : Type of the errors in y: [err, std, none]
                                          default is 'std'.
            xerr (Optional[ndarray])    : Type of the errors in x: [err, std, none]
                                          default is 'std'.
            xgrid (Optional[ndarray])   : Grid to bin the data to -
                                          default in empty so use the
                                          x-axis of the first scan.
            norm2one (Optional[bool])   : Norm transient data to 1 for t < t0
                                          default is False.

        """
        # get scan_sequence data without plotting
        sequence_data, parameters, names, label_texts = self.plot_scan_sequence(
            scan_sequence,
            xgrid=xgrid,
            yerr=yerr,
            xerr=xerr,
            norm2one=norm2one,
            binning=binning,
            skip_plot=True)

        for i, label_text in enumerate(label_texts):
            # travserse the sequence

            header = ''
            saveData = []
            for counter in sequence_data:
                # travserse all counters in the data

                # build the file header
                header = header + counter + '\t '
                # build the data matrix
                saveData.append(sequence_data[counter][i])

            # save data with header to text file
            np.savetxt('{:s}/{:s}_{:s}.dat'.format(path, fileName,
                                                   "".join(x for x in label_text if x.isalnum())),
                       np.r_[saveData].T, delimiter='\t', header=header)

    def fit_scans(self, scans, mod, pars, ylims=[], xlims=[], fig_size=[], xgrid=[],
                  yerr='std', xerr='std', norm2one=False, binning=True,
                  sequence_type='text', label_text='', title_text='', ytext='',
                  xtext='', select='', fit_report=0, show_single=False,
                  weights=False, fit_method='leastsq', offset_t0=False,
                  plot_separate=False, grid_on=True, fmt='o'):
        """Fit, plot, and return the data of scans.

            This is just a wrapper for the fit_scan_sequence method
        """
        scan_sequence = [[scans, '']]
        return self.fit_scan_sequence(scan_sequence, mod, pars, ylims, xlims, fig_size,
                                      xgrid, yerr, xerr, norm2one, binning,
                                      'none', label_text, title_text, ytext,
                                      xtext, select, fit_report, show_single,
                                      weights, fit_method, offset_t0, plot_separate,
                                      grid_on, fmt=fmt)

    def fit_scan_sequence(self, scan_sequence, mod, pars, ylims=[], xlims=[], fig_size=[],
                          xgrid=[], yerr='std', xerr='std', norm2one=False,
                          binning=True, sequence_type='', label_text='',
                          title_text='', ytext='', xtext='', select='',
                          fit_report=0, show_single=False, weights=False,
                          fit_method='leastsq', offset_t0=False,
                          plot_separate=False, grid_on=True,
                          last_res_as_par=False, sequence_data=[], fmt='o'):
        """Fit, plot, and return the data of a scan sequence.

        Args:
            scan_sequence (ndarray[List[int]
                          , int/str])   : Sequence of scan lists and parameters.
            mod (Model[lmfit])          : lmfit model for fitting the data.
            pars (Parameters[lmfit])    : lmfit parameters for fitting the data.
            ylims (Optional[ndarray])   : ylim for the plot.
            xlims (Optional[ndarray])   : xlim for the plot.
            fig_size (Optional[ndarray]) : Figure size of the figure.
            xgrid (Optional[ndarray])   : Grid to bin the data to -
                                          default in empty so use the
                                          x-axis of the first scan.
            yerr (Optional[ndarray])    : Type of the errors in y: [err, std, none]
                                          default is 'std'.
            xerr (Optional[ndarray])    : Type of the errors in x: [err, std, none]
                                          default is 'std'.
            norm2one (Optional[bool])   : Norm transient data to 1 for t < t0
                                          default is False.
            sequence_type (Optional[str]): Type of the sequence: [fluence, delay,
                                          energy, theta] - default is fluence.
            label_text (Optional[str])   : Label of the plot - default is none.
            title_text (Optional[str])   : Title of the figure - default is none.
            ytext (Optional[str])       : y-Label of the plot - defaults is none.
            xtext (Optional[str])       : x-Label of the plot - defaults is none.
            select (Optional[str])      : String to evaluate as select statement
                                          for the fit region - default is none
            fit_report (Optional[int])   : Set the fit reporting level:
                                          [0: none, 1: basic, 2: full]
                                          default 0.
            show_single (Optional[bool]) : Plot each fit seperately - default False.
            weights (Optional[bool])    : Use weights for fitting - default False.
            fit_method (Optional[str])   : Method to use for fitting; refer to
                                          lmfit - default is 'leastsq'.
            offset_t0 (Optional[bool])   : Offset time scans by the fitted
                                          t0 parameter - default False.
            plot_separate (Optional[bool]):A single plot for each counter
                                          default False.
            grid_on (Optional[bool])     : Add grid to plot - default is True.
            last_res_as_par (Optional[bool]): Use the last fit result as start
                                           values for next fit - default is False.
            sequence_data (Optional[ndarray]): actual exp. data are externally given.
                                              default is empty
            fmt (Optional[str])         : format string of the plot - defaults is -o.


        Returns:
            res (Dict[ndarray])        : Fit results.
            parameters (ndarray)       : Parameters of the sequence.
            sequence_data (OrderedDict) : Dictionary of the averaged scan data.equenceData

        """

        # get the last open figure number
        main_fig_num = self.get_last_fig_number()

        if not fig_size:
            # use default figure size if none is given
            fig_size = mpl.rcParams['figure.figsize']

        # initialization of returns
        res = {}  # initialize the results dict

        for i, counter in enumerate(self.get_clist()):
            # traverse all counters in the counter list to initialize the returns

            # results for this counter is again a Dict
            res[counter] = {}

            if isinstance(pars, (list, tuple)):
                # the fit paramters might individual for each counter
                _pars = pars[i]
            else:
                _pars = pars

            for pname, par in _pars.items():
                # add a dict key for each fit parameter in the result dict
                res[counter][pname] = []
                res[counter][pname + 'Err'] = []

            # add some more results
            res[counter]['chisqr'] = []
            res[counter]['redchi'] = []
            res[counter]['CoM'] = []
            res[counter]['int'] = []
            res[counter]['fit'] = []

        if len(sequence_data) > 0:
            # get only the parameters
            _, parameters, names, label_texts = self.plot_scan_sequence(
                scan_sequence,
                ylims=ylims,
                xlims=xlims,
                fig_size=fig_size,
                xgrid=xgrid,
                yerr=yerr,
                xerr=xerr,
                norm2one=norm2one,
                binning=True,
                sequence_type=sequence_type,
                label_text=label_text,
                title_text=title_text,
                skip_plot=True)
        else:
            # get the sequence data and parameters
            sequence_data, parameters, names, label_texts = self.plot_scan_sequence(
                scan_sequence,
                ylims=ylims,
                xlims=xlims,
                fig_size=fig_size,
                xgrid=xgrid,
                yerr=yerr,
                xerr=xerr,
                norm2one=norm2one,
                binning=True,
                sequence_type=sequence_type,
                label_text=label_text,
                title_text=title_text,
                skip_plot=True)

        # this is the number of different counters
        num_sub_plots = len(self.get_clist())

        # fitting and plotting the data
        l_plot = 1  # counter for single plots

        for i, parameter in enumerate(parameters):
            # traverse all parameters of the sequence
            lt = label_texts[i]
            name = names[i]

            x2plot = sequence_data[self.xcol][i]
            xerr2plot = sequence_data[self.xcol + 'Err'][i]

            if fit_report > 0:
                # plot for basics and full fit reporting
                print('')
                print('='*10 + ' Parameter: ' + lt + ' ' + '='*15)

            j = 0  # counter for counters ;)
            k = 1  # counter for subplots
            for counter in sequence_data:
                # traverse all counters in the sequence

                # plot only y counters - next is the coresp. error
                if j >= 2 and j % 2 == 0:

                    # add the counter name to the label for not seperate plots
                    if sequence_type == 'none':
                        _lt = counter
                    else:
                        if plot_separate or num_sub_plots == 1:
                            _lt = lt
                        else:
                            _lt = lt + ' | ' + counter

                    # get the fit models and fit parameters if they are lists/tupels
                    if isinstance(mod, (list, tuple)):
                        _mod = mod[k-1]
                    else:
                        _mod = mod

                    if last_res_as_par and i > 0:
                        # use last results as start values for pars
                        _pars = pars
                        for pname, par in pars.items():
                            _pars[pname].value = res[counter][pname][i-1]
                    else:
                        if isinstance(pars, (list, tuple)):
                            _pars = pars[k-1]
                        else:
                            _pars = pars

                    # get the actual y-data and -errors for plotting and fitting
                    y2plot = sequence_data[counter][i]
                    yerr2plot = sequence_data[counter + 'Err'][i]

                    # evaluate the select statement
                    if select == '':
                        # select all
                        sel = np.ones_like(y2plot, dtype=bool)
                    else:
                        sel = eval(select)

                    # execute the select statement
                    y2plot = y2plot[sel]
                    x2plot = x2plot[sel]
                    yerr2plot = yerr2plot[sel]
                    xerr2plot = xerr2plot[sel]

                    # remove nans
                    y2plot = y2plot[~np.isnan(y2plot)]
                    x2plot = x2plot[~np.isnan(y2plot)]
                    yerr2plot = yerr2plot[~np.isnan(y2plot)]
                    xerr2plot = xerr2plot[~np.isnan(y2plot)]

                    # do the fitting with or without weighting the data
                    if weights:
                        out = _mod.fit(y2plot, _pars, x=x2plot,
                                       weights=1/yerr2plot, method=fit_method,
                                       nan_policy='propagate')
                    else:
                        out = _mod.fit(y2plot, _pars, x=x2plot,
                                       method=fit_method, nan_policy='propagate')

                    if fit_report > 0:
                        # for basic and full fit reporting
                        print('')
                        print('-'*10 + ' ' + counter + ': ' + '-'*15)
                        for key in out.best_values:
                            print('{:>12}:  {:>10.4e} '.format(
                                key, out.best_values[key]))

                    # set the x-offset for delay scans - offset parameter in
                    # the fit must be called 't0'
                    if offset_t0:
                        offsetX = out.best_values['t0']
                    else:
                        offsetX = 0

                    plt.figure(main_fig_num)  # select the main figure

                    if plot_separate:
                        # use subplot for separate plotting
                        plt.subplot((num_sub_plots+num_sub_plots % 2)/2, 2, k)

                    # plot the fit and the data as errorbars
                    x2plotFit = np.linspace(
                        np.min(x2plot), np.max(x2plot), 10000)
                    plot = plt.plot(x2plotFit-offsetX,
                                    out.eval(x=x2plotFit), '-', lw=2, alpha=1)
                    plt.errorbar(x2plot-offsetX, y2plot, fmt=fmt, xerr=xerr2plot,
                                 yerr=yerr2plot, label=_lt, alpha=0.25, color=plot[0].get_color())

                    if len(parameters) > 5:
                        # move the legend outside the plot for more than
                        # 5 sequence parameters
                        plt.legend(bbox_to_anchor=(0., 1.08, 1, .102), frameon=True,
                                   loc=3, numpoints=1, ncol=3, mode="expand",
                                   borderaxespad=0.)
                    else:
                        plt.legend(frameon=True, loc=0, numpoints=1)

                    # set the axis limits, title, labels and gird
                    if xlims:
                        plt.xlim(xlims)
                    if ylims:
                        plt.ylim(ylims)
                    if len(title_text) > 0:
                        if isinstance(title_text, (list, tuple)):
                            plt.title(title_text[k-1])
                        else:
                            plt.title(title_text)
                    else:
                        plt.title(name)

                    if len(xtext) > 0:
                        plt.xlabel(xtext)

                    if len(ytext) > 0:
                        if isinstance(ytext, (list, tuple)):
                            plt.ylabel(ytext[k-1])
                        else:
                            plt.ylabel(ytext)

                    if grid_on:
                        plt.grid(True)

                    # show the single fits and residuals
                    if show_single:
                        plt.figure(main_fig_num+l_plot, figsize=fig_size)
                        gs = mpl.gridspec.GridSpec(
                            2, 1, height_ratios=[1, 3], hspace=0.1)
                        ax1 = plt.subplot(gs[0])
                        markerline, stemlines, baseline = plt.stem(
                            x2plot-offsetX, out.residual, markerfmt=' ',
                            use_line_collection=True)
                        plt.setp(stemlines, 'color',
                                 plot[0].get_color(), 'linewidth', 2, alpha=0.5)
                        plt.setp(baseline, 'color', 'k', 'linewidth', 0)

                        ax1.xaxis.tick_top()
                        ax1.yaxis.set_major_locator(plt.MaxNLocator(3))
                        plt.ylabel('Residuals')
                        if xlims:
                            plt.xlim(xlims)
                        if ylims:
                            plt.ylim(ylims)

                        if len(xtext) > 0:
                            plt.xlabel(xtext)

                        if grid_on:
                            plt.grid(True)

                        if len(title_text) > 0:
                            if isinstance(title_text, (list, tuple)):
                                plt.title(title_text[k-1])
                            else:
                                plt.title(title_text)
                        else:
                            plt.title(name)
                        ax2 = plt.subplot(gs[1])
                        x2plotFit = np.linspace(
                            np.min(x2plot), np.max(x2plot), 1000)
                        ax2.plot(x2plotFit-offsetX, out.eval(x=x2plotFit),
                                 '-', lw=2, alpha=1, color=plot[0].get_color())
                        ax2.errorbar(x2plot-offsetX, y2plot, fmt=fmt, xerr=xerr2plot,
                                     yerr=yerr2plot, label=_lt, alpha=0.25,
                                     color=plot[0].get_color())
                        plt.legend(frameon=True, loc=0, numpoints=1)

                        if xlims:
                            plt.xlim(xlims)
                        if ylims:
                            plt.ylim(ylims)

                        if len(xtext) > 0:
                            plt.xlabel(xtext)

                        if len(ytext) > 0:
                            if isinstance(ytext, (list, tuple)):
                                plt.ylabel(ytext[k-1])
                            else:
                                plt.ylabel(ytext)

                        if grid_on:
                            plt.grid(True)

                        l_plot += 1
                    if fit_report > 1:
                        # for full fit reporting
                        print('_'*40)
                        print(out.fit_report())

                    # add the fit results to the returns
                    for pname, par in _pars.items():
                        res[counter][pname] = np.append(
                            res[counter][pname], out.best_values[pname])
                        res[counter][pname + 'Err'] = np.append(
                            res[counter][pname + 'Err'], out.params[pname].stderr)

                    res[counter]['chisqr'] = np.append(
                        res[counter]['chisqr'], out.chisqr)
                    res[counter]['redchi'] = np.append(
                        res[counter]['redchi'], out.redchi)
                    res[counter]['CoM'] = np.append(
                        res[counter]['CoM'], sum(y2plot*x2plot)/sum(y2plot))
                    res[counter]['int'] = np.append(
                        res[counter]['int'], sum(y2plot))
                    res[counter]['fit'] = np.append(res[counter]['fit'], out)

                    k += 1

                j += 1

        plt.figure(main_fig_num)  # set as active figure

        return res, parameters, sequence_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2021 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

from .. import config
import logging

__all__ = ['Scan']

__docformat__ = 'restructuredtext'

import numpy as np


class Scan(object):
    """Scan

    Class for scan which hold the relevant (meta) data.

    Args:
        number (uint): number of the scan.

    Keyword Args:
        cmd (str): scan command.
        user (str): scan user.
        date (str): scan date.
        time (str): scan time.
        int_time (float): integration time.
        init_mopo (dict(float)): initial motor position.
        header (str): full scan header.

    Attributes:
        log (logging.logger): logger instance from logging.
        number (uint): number of the scan.
        meta (dict): meta data dictionary.
        data (ndarray[float]): all data recarray

    """

    def __init__(self, number, **kwargs):
        self.log = logging.getLogger(__name__)
        self.log.setLevel(config.LOG_LEVEL)
        self.log.debug('Creating scan #{:d}'.format(number))
        self.number = np.uint64(number)
        # initialize empty data array and circumvent
        # check for recarray here
        self._data = None
        self.scalar_data_names = []
        self.oned_data_names = []
        self.twod_data_names = []
        self.index_data()
        self.meta = {}
        self.meta['number'] = self.number
        self.meta['cmd'] = kwargs.get('cmd', '')
        self.meta['user'] = kwargs.get('user', '')
        self.meta['date'] = kwargs.get('date', '')
        self.meta['time'] = kwargs.get('time', '')
        self.meta['int_time'] = kwargs.get('int_time', '')
        self.meta['init_mopo'] = kwargs.get('init_mopo', {})
        self.meta['header'] = kwargs.get('header', '')

    def __getattr__(self, attr):
        """__getattr__

        Allows to access the data and meta(init_mopo) keys as scan attributes.

        Returns:
            attr (ndarray[float]|float|str): data/meta values.

        """
        # check data recarray
        try:
            return self.data[attr]
        except (ValueError, IndexError, TypeError):
            pass

        # check meta dict
        try:
            return self.meta[attr]
        except KeyError:
            pass

        # check meta init_mopo dict
        try:
            return self.meta['init_mopo'][attr]
        except KeyError:
            raise AttributeError('Scan has no attribute {:s}'.format(attr))

    def index_data(self):
        """index_data

        Check the dimensions of the data recarray elements and
        remember the names for scaler, 1d, and 2d data columns.

        """
        if self.data is not None:
            for descr in self.data.dtype.descr:
                try:
                    if len(descr[2]) == 1:
                        self.oned_data_names.append(descr[0])
                    elif len(descr[2]) == 2:
                        self.twod_data_names.append(descr[0])
                except Exception:
                    self.scalar_data_names.append(descr[0])

    def get_scalar_data(self):
        """get_scalar_data

        Returns only scalar data from the data recarray.

        Returns:
            data (ndarray[float]): scalar data.

        """
        if self.scalar_data_names == []:
            return None
        else:
            return self.data[self.scalar_data_names]

    def get_oned_data(self):
        """get_oned_data

        Returns only scalar data from the data recarray.

        Returns:
            data (ndarray[float]): 1d data.

        """
        if self.oned_data_names == []:
            return None
        else:
            return self.data[self.oned_data_names]

    def get_twod_data(self):
        """get_twod_data

        Returns only scalar data from the data recarray.

        Returns:
            data (ndarray[float]): 2d data.

        """
        if self.twod_data_names == []:
            return None
        else:
            return self.data[self.twod_data_names]

    def clear_data(self):
        self._data = None
        self.log.debug('Cleared data for scan #{:d}'.format(self.number))

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if isinstance(data, np.recarray):
            self._data = data
        elif data is None:
            self.log.info('Scan #{:d} contains no data!'.format(self.number))
            self._data = None
        else:
            raise TypeError('data must be numpy recarray')



from setuptools import setup, find_packages

setup(
    name='pyEvalData',
    version='1.3.7',
    packages=find_packages(),
    url='https://github.com/dschick/pyEvalData',
    install_requires=['numpy',
                      'matplotlib',
                      'lmfit',
                      'scipy',
                      'uncertainties',
                      'xrayutilities',
                      'h5py>=3.0',
                      'nexusformat'],
    extras_require={
        'testing': ['flake8', 'pytest'],
        'documentation': ['sphinx', 'nbsphinx', 'sphinxcontrib-napoleon'],
    },
    license='MIT',
    author='Daniel Schick',
    author_email='schick.daniel@gmail.com',
    description='Python Modul to evaluate experimental data',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    package_data={
        'pyEvalData': ['*.conf']
    },
    python_requires='>=3.5',
    keywords='data evaluation',
)



from .source import Source
from .scan import Scan
from .spec import Spec
from .sardana_nexus import SardanaNeXus
from .palxfel import PalH5

__all__ = ['Source', 'Scan', 'Spec', 'SardanaNeXus', 'PalH5']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2015-2021 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.


from numpy.core.records import fromarrays
import nexusformat.nexus as nxs
import os.path as path

from .source import Source
from .scan import Scan


class SardanaNeXus(Source):
    """SardanaNeXus

    Source implementation for Sardana NeXus files.

    Args:
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.

    Keyword Args:
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        read_all_data (bool): read all data on parsing.
          If false, data will be read only on demand.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    Attributes:
        log (logging.logger): logger instance from logging.
        name (str): name of the source
        scan_dict (dict(scan)): dict of scan objects with
          key being the scan number.
        start_scan_number (uint): start of scan numbers to parse.
        stop_scan_number (uint): stop of scan numbers to parse.
          This number is included.
        file_name (str): file name including extension,
          can include regex pattern.
        file_path (str): file path.
        h5_file_name (str): name for generated h5 file.
        h5_file_name_postfix (str): postfix for h5 file name.
        h5_file_path (str): path for generated h5 file.
        h5_file_exists(bool): if h5 file exists.
        read_all_data (bool): read all data on parsing.
        read_and_forget (bool): clear data after read to save memory.
        update_before_read (bool): always update from source
          before reading scan data.
        use_h5 (bool): use h5 file to join/compress raw data.
        force_overwrite (bool): forced re-read of raw source and
          re-generated of h5 file.

    """
    def __init__(self, file_name, file_path, **kwargs):
        super().__init__(file_name, file_path, **kwargs)

    def parse_raw(self):
        """parse_raw

        Parse the Sardana NeXus file and populate the `scan_dict`.

        """
        self.log.info('parse_raw')
        nxs_file_path = path.join(self.file_path, self.file_name)
        try:
            nxs_file = nxs.nxload(nxs_file_path, mode='r')
        except nxs.NeXusError:
            raise nxs.NeXusError('Sardana NeXus file \'{:s}\' does not exist!'.format(
                nxs_file_path))

        with nxs_file.nxfile:
            for entry in nxs_file:
                # check for scan number in given range
                entry_number = int(nxs_file[entry].entry_identifier)
                if (entry_number >= self.start_scan_number) and \
                        ((entry_number <= self.stop_scan_number) or
                            (self.stop_scan_number == -1)):
                    last_scan_number = self.get_last_scan_number()
                    # check if Scan needs to be re-created
                    # if scan is not present, its the last one, or force overwrite
                    if (entry_number not in self.scan_dict.keys()) or \
                            (entry_number >= last_scan_number) or \
                            self.force_overwrite:
                        # create scan object
                        init_mopo = {}
                        for field in nxs_file[entry].measurement.pre_scan_snapshot:
                            init_mopo[field] = \
                                nxs_file[entry]['measurement/pre_scan_snapshot'][field]

                        scan = Scan(int(entry_number),
                                    cmd=nxs_file[entry].title,
                                    date=nxs_file[entry].start_time,
                                    time=nxs_file[entry].start_time,
                                    int_time=float(0),
                                    header='',
                                    init_mopo=init_mopo)
                        self.scan_dict[entry_number] = scan
                        # check if the data needs to be read as well
                        if self.read_all_data:
                            self.read_scan_data(self.scan_dict[entry_number])

    def read_raw_scan_data(self, scan):
        """read_raw_scan_data

        Reads the data for a given scan object from Sardana NeXus file.

        Args:
            scan (Scan): scan object.

        """
        self.log.info('read_raw_scan_data for scan #{:d}'.format(scan.number))
        # try to open the file
        nxs_file_path = path.join(self.file_path, self.file_name)
        try:
            nxs_file = nxs.nxload(nxs_file_path, mode='r')
        except nxs.NeXusError:
            raise nxs.NeXusError('Sardana NeXus file \'{:s}\' does not exist!'.format(
                nxs_file_path))
        entry_name = 'entry{:d}'.format(scan.number)
        # try to enter entry
        try:
            entry = nxs_file[entry_name]
        except nxs.NeXusError:
            self.log.exception('Entry #{:d} not present in NeXus file!'.format(scan.number))
            return
        # iterate through data fields
        data_list = []
        dtype_list = []
        for field in entry.measurement:
            # do not add data which is already in the pre-scan snapshot
            # that is tricky if it is in the snapshot and scanned ...
            if field != 'pre_scan_snapshot':
                data_list.append(entry.measurement[field])
                dtype_list.append((field,
                                   entry.measurement[field].dtype,
                                   entry.measurement[field].shape))
        if len(data_list) > 0:
            scan.data = fromarrays(data_list, dtype=dtype_list)
        else:
            scan.date = None
